# coding: utf-8
"frozen and hidden values"
import autopath
from py.test import raises

from tiramisu.config import *
from tiramisu.option import *

def make_description():
    gcoption = ChoiceOption('name', 'GC name', ('ref', 'framework'), 'ref')
    gcdummy = BoolOption('dummy', 'dummy', default=False, properties=(('hidden'),))
    objspaceoption = ChoiceOption('objspace', 'Object space',
                                ('std', 'thunk'), 'std')
    booloption = BoolOption('bool', 'Test boolean option', default=True)
    intoption = IntOption('int', 'Test int option', default=0)
    floatoption = FloatOption('float', 'Test float option', default=2.3)
    stroption = StrOption('str', 'Test string option', default="abc")

    wantref_option = BoolOption('wantref', 'Test requires', default=False,
                                    requires=(('gc.name', 'ref', 'hidden'),))
    wantframework_option = BoolOption('wantframework', 'Test requires',
                                      default=False,
                                      requires=(('gc.name', 'framework', 'hidden'),))

    # ____________________________________________________________
    booloptiontwo = BoolOption('booltwo', 'Test boolean option two', default=False)
    subgroup = OptionDescription('subgroup', '', [booloptiontwo])
    # ____________________________________________________________

    gcgroup = OptionDescription('gc', '', [subgroup, gcoption, gcdummy, floatoption])
    descr = OptionDescription('trs', '', [gcgroup, booloption, objspaceoption,
                                           wantref_option, stroption,
                                           wantframework_option,
                                           intoption])
    return descr
#____________________________________________________________
#freeze
def make_description_freeze():
    gcoption = ChoiceOption('name', 'GC name', ('ref', 'framework'), 'ref')
    gcdummy = BoolOption('dummy', 'dummy', default=False)
    objspaceoption = ChoiceOption('objspace', 'Object space',
                                ('std', 'thunk'), 'std')
    booloption = BoolOption('bool', 'Test boolean option', default=True)
    intoption = IntOption('int', 'Test int option', default=0)
    floatoption = FloatOption('float', 'Test float option', default=2.3)
    stroption = StrOption('str', 'Test string option', default="abc")
    boolop = BoolOption('boolop', 'Test boolean option op', default=True)
    wantref_option = BoolOption('wantref', 'Test requires', default=False,
                                    requires=(('boolop', True, 'hidden'),))
    wantframework_option = BoolOption('wantframework', 'Test requires',
                                      default=False,
                                      requires=(('boolop', True, 'hidden'),))

    gcgroup = OptionDescription('gc', '', [gcoption, gcdummy, floatoption])
    descr = OptionDescription('tiramisu', '', [gcgroup, booloption, objspaceoption,
                                           wantref_option, stroption,
                                           wantframework_option,
                                           intoption, boolop])
    return descr

def test_freeze_whole_config():
    descr = make_description_freeze()
    conf = Config(descr)
    settings = conf.cfgimpl_get_settings()
    settings.enable_property('everything_frozen')
    assert conf.gc.dummy == False
    raises(TypeError, "conf.gc.dummy = True")
    settings.disable_property('everything_frozen')
    conf.gc.dummy = True
    assert conf.gc.dummy == True

def test_freeze_one_option():
    "freeze an option "
    descr = make_description_freeze()
    conf = Config(descr)
    setting = conf.cfgimpl_get_settings()
    setting.read_write()
    #freeze only one option
    dummy = conf.unwrap_from_path('gc.dummy')
    conf.gc.cfgimpl_get_settings().add_property('frozen', dummy)
    assert conf.gc.dummy == False
    raises(TypeError, "conf.gc.dummy = True")

def test_frozen_value():
    "setattr a frozen value at the config level"
    s = StrOption("string", "", default="string")
    descr = OptionDescription("options", "", [s])
    config = Config(descr)
    settings = config.cfgimpl_get_settings().enable_property('frozen')
    config.cfgimpl_get_settings().add_property('frozen', s)
    raises(TypeError, 'config.string = "egg"')

def test_freeze():
    "freeze a whole configuration object"
    descr = make_description()
    conf = Config(descr)
    settings = conf.cfgimpl_get_settings().enable_property('frozen')
    name = conf.unwrap_from_path("gc.name")
    conf.cfgimpl_get_settings().add_property('frozen', name)
    raises(TypeError, "conf.gc.name = 'framework'")
# ____________________________________________________________
def test_is_hidden():
    descr = make_description()
    config = Config(descr)
    setting = config.cfgimpl_get_settings()
    setting.read_write()
    dummy = config.unwrap_from_path('gc.dummy')
    assert not config.cfgimpl_get_settings().has_property('frozen', dummy)
    # setattr
    raises(PropertiesOptionError, "config.gc.dummy == False")
    # getattr
    raises(PropertiesOptionError, "config.gc.dummy")
    # I want to access to this option anyway
    opt = config.unwrap_from_path("gc.dummy")
    assert config.cfgimpl_get_values()[opt] == False

def test_group_is_hidden():
    descr = make_description()
    config = Config(descr)
    setting = config.cfgimpl_get_settings()
    setting.read_write()
    gc = config.unwrap_from_path('gc')
    dummy = config.unwrap_from_path('gc.dummy')
    config.cfgimpl_get_settings().add_property('hidden', gc)
    raises(PropertiesOptionError, "config.gc.dummy")
    assert config.cfgimpl_get_settings().has_property('hidden', gc)
    raises(PropertiesOptionError, "config.gc.float")
    # manually set the subconfigs to "show"
    config.cfgimpl_get_settings().del_property('hidden', gc)
    assert not config.cfgimpl_get_settings().has_property('hidden', gc)
    assert config.gc.float == 2.3
    #dummy est en hide
    raises(PropertiesOptionError, "config.gc.dummy == False")

def test_global_show():
    descr = make_description()
    config = Config(descr)
    setting = config.cfgimpl_get_settings()
    setting.read_write()
    dummy = config.unwrap_from_path('gc.dummy')
    config.cfgimpl_get_settings().add_property('hidden', dummy)
    assert config.cfgimpl_get_settings().has_property('hidden', dummy)
    raises(PropertiesOptionError, "config.gc.dummy == False")

def test_with_many_subgroups():
    descr = make_description()
    config = Config(descr)
    booltwo = config.unwrap_from_path('gc.subgroup.booltwo')
    assert not config.cfgimpl_get_settings().has_property('hidden', booltwo)
    assert config.gc.subgroup.booltwo == False
    config.cfgimpl_get_settings().add_property('hidden', booltwo)
    path = 'gc.subgroup.booltwo'
    homeconfig, name = config.cfgimpl_get_home_by_path(path)
    assert name == "booltwo"
    option = getattr(homeconfig._cfgimpl_descr, name)
    assert config.cfgimpl_get_settings().has_property('hidden', booltwo)
